@@ -1,14 +1,55 @@
 use std::borrow::Cow;
 use std::collections::{HashMap, HashSet};
 
+use ouroboros::self_referencing;
 use pyo3::exceptions::PyTypeError;
 use pyo3::prelude::*;
 use pyo3::types::{PyString, PyTuple};
 
-/// Sanitize an HTML fragment according to the given options.
+struct Config {
+    tags: Option<HashSet<String>>,
+    clean_content_tags: Option<HashSet<String>>,
+    attributes: Option<HashMap<String, HashSet<String>>>,
+    attribute_filter: Option<PyObject>,
+    strip_comments: bool,
+    link_rel: Option<String>,
+    generic_attribute_prefixes: Option<HashSet<String>>,
+    tag_attribute_values: Option<HashMap<String, HashMap<String, HashSet<String>>>>,
+    set_tag_attribute_values: Option<HashMap<String, HashMap<String, String>>>,
+    url_schemes: Option<HashSet<String>>,
+    allowed_classes: Option<HashMap<String, HashSet<String>>>,
+    filter_style_properties: Option<HashSet<String>>,
+}
+
+impl Default for Config {
+    fn default() -> Self {
+        Self {
+            tags: None,
+            clean_content_tags: None,
+            attributes: None,
+            attribute_filter: None,
+            strip_comments: true,
+            link_rel: Some("noopener noreferrer".to_string()),
+            generic_attribute_prefixes: None,
+            tag_attribute_values: None,
+            set_tag_attribute_values: None,
+            url_schemes: None,
+            allowed_classes: None,
+            filter_style_properties: None,
+        }
+    }
+}
+
+#[self_referencing]
+struct Inner {
+    config: Config,
+    #[borrows(config)]
+    #[not_covariant]
+    builder: ammonia::Builder<'this>,
+}
+
+/// Create a reusable sanitizer according to the given options.
 ///
-/// :param html: Input HTML fragment
-/// :type html: ``str``
 /// :param tags: Sets the tags that are allowed.
 /// :type tags: ``set[str]``, optional
 /// :param clean_content_tags: Sets the tags whose contents will be completely removed from the output.
@@ -52,6 +93,229 @@ use pyo3::types::{PyString, PyTuple};
 ///     invalid declarations and @rules will be removed, with only syntactically valid
 ///     declarations kept.
 /// :type filter_style_properties: ``set[str]``, optional
+#[pyclass]
+pub struct Cleaner {
+    inner: Inner,
+}
+
+impl Cleaner {
+    fn new(config: Config) -> Self {
+        let inner = InnerBuilder {
+            config,
+            builder_builder: |config| Self::build_ammonia_from_config(config),
+        }
+        .build();
+        Self { inner }
+    }
+
+    fn build_ammonia_from_config(config: &Config) -> ammonia::Builder<'_> {
+        let mut builder = ammonia::Builder::default();
+
+        if let Some(tags) = config.tags.as_ref() {
+            let tags: HashSet<&str> = tags.iter().map(|s| s.as_str()).collect();
+            builder.tags(tags);
+        }
+        if let Some(tags) = config.clean_content_tags.as_ref() {
+            let tags: HashSet<&str> = tags.iter().map(|s| s.as_str()).collect();
+            builder.clean_content_tags(tags);
+        }
+        if let Some(attrs) = config.attributes.as_ref() {
+            let attrs: HashMap<&str, HashSet<&str>> = attrs
+                .iter()
+                .filter(|(k, _)| k.as_str() != "*")
+                .map(|(k, v)| (k.as_str(), v.iter().map(|s| s.as_str()).collect()))
+                .collect();
+            builder.tag_attributes(attrs);
+            if let Some(generic_attrs) = config.attributes.as_ref().and_then(|a| a.get("*")) {
+                let generic_attrs: HashSet<&str> =
+                    generic_attrs.iter().map(|s| s.as_str()).collect();
+                builder.generic_attributes(generic_attrs);
+            }
+        }
+        if let Some(prefixes) = config.generic_attribute_prefixes.as_ref() {
+            let prefixes: HashSet<&str> = prefixes.iter().map(|s| s.as_str()).collect();
+            builder.generic_attribute_prefixes(prefixes);
+        }
+        if let Some(values) = config.tag_attribute_values.as_ref() {
+            let values: HashMap<&str, HashMap<&str, HashSet<&str>>> = values
+                .iter()
+                .map(|(tag, attrs)| {
+                    let inner: HashMap<&str, HashSet<&str>> = attrs
+                        .iter()
+                        .map(|(attr, vals)| {
+                            (attr.as_str(), vals.iter().map(|v| v.as_str()).collect())
+                        })
+                        .collect();
+                    (tag.as_str(), inner)
+                })
+                .collect();
+            builder.tag_attribute_values(values);
+        }
+        if let Some(values) = config.set_tag_attribute_values.as_ref() {
+            let values: HashMap<&str, HashMap<&str, &str>> = values
+                .iter()
+                .map(|(tag, attrs)| {
+                    let inner: HashMap<&str, &str> = attrs
+                        .iter()
+                        .map(|(attr, val)| (attr.as_str(), val.as_str()))
+                        .collect();
+                    (tag.as_str(), inner)
+                })
+                .collect();
+            builder.set_tag_attribute_values(values);
+        }
+        let attribute_filter = config
+            .attribute_filter
+            .as_ref()
+            .map(|f| Python::with_gil(|py| f.clone_ref(py)));
+        if let Some(callback) = attribute_filter {
+            builder.attribute_filter(move |element, attribute, value| {
+                Python::with_gil(|py| {
+                    let res = callback.call(
+                        py,
+                        PyTuple::new(
+                            py,
+                            [
+                                PyString::new(py, element),
+                                PyString::new(py, attribute),
+                                PyString::new(py, value),
+                            ],
+                        )
+                        .unwrap(),
+                        None,
+                    );
+                    let err = match res {
+                        Ok(val) => {
+                            if val.is_none(py) {
+                                return None;
+                            } else if let Ok(s) = val.extract::<String>(py) {
+                                return Some(Cow::<str>::Owned(s));
+                            } else {
+                                PyTypeError::new_err(
+                                    "expected attribute_filter to return str or None",
+                                )
+                            }
+                        }
+                        Err(err) => err,
+                    };
+                    err.write_unraisable(
+                        py,
+                        Some(
+                            &PyTuple::new(
+                                py,
+                                [
+                                    PyString::new(py, element),
+                                    PyString::new(py, attribute),
+                                    PyString::new(py, value),
+                                ],
+                            )
+                            .unwrap(),
+                        ),
+                    );
+                    Some(value.into())
+                })
+            });
+        }
+        builder.strip_comments(config.strip_comments);
+        builder.link_rel(config.link_rel.as_deref());
+        if let Some(url_schemes) = config.url_schemes.as_ref() {
+            let url_schemes: HashSet<_> = url_schemes.iter().map(|s| s.as_str()).collect();
+            builder.url_schemes(url_schemes);
+        }
+        if let Some(allowed_classes) = config.allowed_classes.as_ref() {
+            builder.allowed_classes(
+                allowed_classes
+                    .iter()
+                    .map(|(tag, class_set)| {
+                        (tag.as_str(), class_set.iter().map(|c| c.as_str()).collect())
+                    })
+                    .collect(),
+            );
+        }
+        if let Some(filter_style_properties) = config.filter_style_properties.as_ref() {
+            builder.filter_style_properties(
+                filter_style_properties
+                    .iter()
+                    .map(|prop| prop.as_str())
+                    .collect(),
+            );
+        }
+
+        builder
+    }
+
+    pub fn clean(&self, html: &str) -> String {
+        self.inner
+            .with_builder(|builder| builder.clean(html).to_string())
+    }
+}
+
+#[pymethods]
+impl Cleaner {
+    #[new]
+    #[pyo3(signature = (
+        tags = None,
+        clean_content_tags = None,
+        attributes = None,
+        attribute_filter = None,
+        strip_comments = true,
+        link_rel = "noopener noreferrer",
+        generic_attribute_prefixes = None,
+        tag_attribute_values = None,
+        set_tag_attribute_values = None,
+        url_schemes = None,
+        allowed_classes = None,
+        filter_style_properties = None
+    ))]
+    fn py_new(
+        py: Python,
+        tags: Option<HashSet<String>>,
+        clean_content_tags: Option<HashSet<String>>,
+        attributes: Option<HashMap<String, HashSet<String>>>,
+        attribute_filter: Option<PyObject>,
+        strip_comments: bool,
+        link_rel: Option<&str>,
+        generic_attribute_prefixes: Option<HashSet<String>>,
+        tag_attribute_values: Option<HashMap<String, HashMap<String, HashSet<String>>>>,
+        set_tag_attribute_values: Option<HashMap<String, HashMap<String, String>>>,
+        url_schemes: Option<HashSet<String>>,
+        allowed_classes: Option<HashMap<String, HashSet<String>>>,
+        filter_style_properties: Option<HashSet<String>>,
+    ) -> PyResult<Self> {
+        if let Some(callback) = attribute_filter.as_ref() {
+            if !callback.bind(py).is_callable() {
+                return Err(PyTypeError::new_err("attribute_filter must be callable"));
+            }
+        }
+        let config = Config {
+            tags,
+            clean_content_tags,
+            attributes,
+            attribute_filter,
+            strip_comments,
+            link_rel: link_rel.map(|s| s.to_string()),
+            generic_attribute_prefixes,
+            tag_attribute_values,
+            set_tag_attribute_values,
+            url_schemes,
+            allowed_classes,
+            filter_style_properties,
+        };
+        Ok(Self::new(config))
+    }
+
+    /// Sanitize an HTML fragment
+    #[pyo3(name = "clean")]
+    fn py_clean(&self, py: Python, html: &str) -> PyResult<String> {
+        Ok(py.allow_threads(|| self.clean(html)))
+    }
+}
+
+/// Sanitize an HTML fragment according to the given options.
+/// See ``Cleaner()`` for detailed sanitizer options.
+///
+/// :param html: Input HTML fragment
+/// :type html: ``str``
 /// :return: Sanitized HTML fragment
 /// :rtype: ``str``
 ///
@@ -68,7 +332,7 @@ use pyo3::types::{PyString, PyTuple};
 /// Example of using ``attribute_filter``:
 ///
 /// .. code-block:: pycon
-///   
+///
 ///    >>> from copy import deepcopy
 ///    >>> attributes = deepcopy(nh3.ALLOWED_ATTRIBUTES)
 ///    >>> attributes["a"].add("class")
@@ -78,7 +342,7 @@ use pyo3::types::{PyString, PyTuple};
 ///    ...             return "mention"
 ///    ...         return None
 ///    ...     return value
-///    >>> nh3.clean("<a class='mention unwanted'>@foo</a>", 
+///    >>> nh3.clean("<a class='mention unwanted'>@foo</a>",
 ///    ...     attributes=attributes,
 ///    ...     attribute_filter=attribute_filter)
 ///    '<a class="mention" rel="noopener noreferrer">@foo</a>'
@@ -94,8 +358,6 @@ use pyo3::types::{PyString, PyTuple};
 ///    ...     link_rel=None, attributes=attributes)
 ///    '<a href="https://tag.example" rel="tag">#tag</a>'
 
-
-
 #[pyfunction(signature = (
     html,
     tags = None,
@@ -117,7 +379,7 @@ fn clean(
     html: &str,
     tags: Option<HashSet<String>>,
     clean_content_tags: Option<HashSet<String>>,
-    mut attributes: Option<HashMap<String, HashSet<String>>>,
+    attributes: Option<HashMap<String, HashSet<String>>>,
     attribute_filter: Option<PyObject>,
     strip_comments: bool,
     link_rel: Option<&str>,
@@ -126,151 +388,24 @@ fn clean(
     set_tag_attribute_values: Option<HashMap<String, HashMap<String, String>>>,
     url_schemes: Option<HashSet<String>>,
     allowed_classes: Option<HashMap<String, HashSet<String>>>,
-    filter_style_properties: Option<HashSet<String>>
+    filter_style_properties: Option<HashSet<String>>,
 ) -> PyResult<String> {
-    if let Some(callback) = attribute_filter.as_ref() {
-        if !callback.bind(py).is_callable() {
-            return Err(PyTypeError::new_err("attribute_filter must be callable"));
-        }
-    }
-
-    let cleaned = py.allow_threads(|| {
-        if tags.is_some()
-            || clean_content_tags.is_some()
-            || attributes.is_some()
-            || attribute_filter.is_some()
-            || !strip_comments
-            || link_rel.as_deref() != Some("noopener noreferrer")
-            || generic_attribute_prefixes.is_some()
-            || tag_attribute_values.is_some()
-            || set_tag_attribute_values.is_some()
-            || url_schemes.is_some()
-            || allowed_classes.is_some()
-            || filter_style_properties.is_some()
-        {
-            let generic_attrs = attributes.as_mut().and_then(|attrs| attrs.remove("*"));
-            let mut cleaner = ammonia::Builder::default();
-            if let Some(tags) = tags.as_ref() {
-                let tags: HashSet<&str> = tags.iter().map(|s| s.as_str()).collect();
-                cleaner.tags(tags);
-            }
-            if let Some(tags) = clean_content_tags.as_ref() {
-                let tags: HashSet<&str> = tags.iter().map(|s| s.as_str()).collect();
-                cleaner.clean_content_tags(tags);
-            }
-            if let Some(attrs) = attributes.as_ref() {
-                let attrs: HashMap<&str, HashSet<&str>> = attrs
-                    .iter()
-                    .map(|(k, v)| (k.as_str(), v.iter().map(|s| s.as_str()).collect()))
-                    .collect();
-                cleaner.tag_attributes(attrs);
-                if let Some(generic_attrs) = generic_attrs.as_ref() {
-                    let generic_attrs: HashSet<&str> =
-                        generic_attrs.iter().map(|s| s.as_str()).collect();
-                    cleaner.generic_attributes(generic_attrs);
-                }
-            }
-            if let Some(prefixes) = generic_attribute_prefixes.as_ref() {
-                let prefixes: HashSet<&str> = prefixes.iter().map(|s| s.as_str()).collect();
-                cleaner.generic_attribute_prefixes(prefixes);
-            }
-            if let Some(values) = tag_attribute_values.as_ref() {
-                let values: HashMap<&str, HashMap<&str, HashSet<&str>>> = values
-                    .iter()
-                    .map(|(tag, attrs)| {
-                        let inner: HashMap<&str, HashSet<&str>> = attrs
-                            .iter()
-                            .map(|(attr, vals)| {
-                                (attr.as_str(), vals.iter().map(|v| v.as_str()).collect())
-                            })
-                            .collect();
-                        (tag.as_str(), inner)
-                    })
-                    .collect();
-                cleaner.tag_attribute_values(values);
-            }
-            if let Some(values) = set_tag_attribute_values.as_ref() {
-                let values: HashMap<&str, HashMap<&str, &str>> = values
-                    .iter()
-                    .map(|(tag, attrs)| {
-                        let inner: HashMap<&str, &str> = attrs
-                            .iter()
-                            .map(|(attr, val)| (attr.as_str(), val.as_str()))
-                            .collect();
-                        (tag.as_str(), inner)
-                    })
-                    .collect();
-                cleaner.set_tag_attribute_values(values);
-            }
-            if let Some(callback) = attribute_filter {
-                cleaner.attribute_filter(move |element, attribute, value| {
-                    Python::with_gil(|py| {
-                        let res = callback.call(
-                            py,
-                            PyTuple::new(
-                                py,
-                                [
-                                    PyString::new(py, element),
-                                    PyString::new(py, attribute),
-                                    PyString::new(py, value),
-                                ],
-                            )
-                            .unwrap(),
-                            None,
-                        );
-                        let err = match res {
-                            Ok(val) => {
-                                if val.is_none(py) {
-                                    return None;
-                                } else if let Ok(s) = val.extract::<String>(py) {
-                                    return Some(Cow::<str>::Owned(s));
-                                } else {
-                                    PyTypeError::new_err(
-                                        "expected attribute_filter to return str or None",
-                                    )
-                                }
-                            }
-                            Err(err) => err,
-                        };
-                        err.write_unraisable(
-                            py,
-                            Some(
-                                &PyTuple::new(
-                                    py,
-                                    [
-                                        PyString::new(py, element),
-                                        PyString::new(py, attribute),
-                                        PyString::new(py, value),
-                                    ],
-                                )
-                                .unwrap(),
-                            ),
-                        );
-                        Some(value.into())
-                    })
-                });
-            }
-            cleaner.strip_comments(strip_comments);
-            cleaner.link_rel(link_rel.as_deref());
-            if let Some(url_schemes) = url_schemes.as_ref() {
-                let url_schemes: HashSet<_> = url_schemes.iter().map(|s| s.as_str()).collect();
-                cleaner.url_schemes(url_schemes);
-            }
-            if let Some(allowed_classes) = allowed_classes.as_ref() {
-                cleaner.allowed_classes(allowed_classes.iter().map(|(tag, class_set)|
-                    (tag.as_str(), class_set.iter().map(|c| c.as_str()).collect())
-                ).collect());
-            }
-            if let Some(filter_style_properties) = filter_style_properties.as_ref() {
-                cleaner.filter_style_properties(filter_style_properties.iter().map(|prop| prop.as_str()).collect());
-            }
-            cleaner.clean(html).to_string()
-        } else {
-            ammonia::clean(html)
-        }
-    });
-
-    Ok(cleaned)
+    let cleaner = Cleaner::py_new(
+        py,
+        tags,
+        clean_content_tags,
+        attributes,
+        attribute_filter,
+        strip_comments,
+        link_rel,
+        generic_attribute_prefixes,
+        tag_attribute_values,
+        set_tag_attribute_values,
+        url_schemes,
+        allowed_classes,
+        filter_style_properties,
+    )?;
+    Ok(py.allow_threads(|| cleaner.clean(html)))
 }
 
 /// Turn an arbitrary string into unformatted HTML.
@@ -327,6 +462,7 @@ fn nh3(_py: Python, m: &Bound<PyModule>) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(clean, m)?)?;
     m.add_function(wrap_pyfunction!(clean_text, m)?)?;
     m.add_function(wrap_pyfunction!(is_html, m)?)?;
+    m.add_class::<Cleaner>()?;
 
     let a = ammonia::Builder::default();
     m.add("ALLOWED_TAGS", a.clone_tags())?;
